@@ -1,17 +1,15 @@
+import asyncio
 import json
 import os
 import string
-from concurrent.futures import ThreadPoolExecutor
 
 import escapism
-import kubernetes.config
 from jupyterhub.proxy import Proxy
 from jupyterhub.utils import exponential_backoff
-from kubernetes import client
-from tornado import gen
-from tornado.concurrent import run_on_executor
+from kubernetes_asyncio import client
 from traitlets import Unicode
 
+from .clients import load_config
 from .clients import shared_client
 from .objects import make_ingress
 from .reflector import ResourceReflector
@@ -44,6 +42,91 @@ class EndpointsReflector(ResourceReflector):
 
 
 class KubeIngressProxy(Proxy):
+    """
+    DISCLAIMER:
+
+        This JupyterHub Proxy class is not maintained thoroughly with tests and
+        documentation, or actively used in any official distribution of
+        JupyterHub.
+
+        KubeIngressProxy was originally developed by @yuvipanda in preparation
+        to support a JupyterHub that could end up with ~20k simultaneous users.
+        The single ConfigurableHTTPProxy server that is controlled by the
+        default JupyterHub Proxy class could bottleneck for such heavy load,
+        even though it has been found to be good enough for ~2k simultaneous
+        users. By instead using this JupyterHub Proxy class that would create
+        k8s resources (Ingress, Service, Endpoint) that in turn controlled how a
+        flexible set of proxy servers would proxy traffic, the bottleneck could
+        be removed.
+
+        KubeIngressProxy's efficiency relates greatly to the performance of the
+        k8s api-server and the k8s controller that routes traffic based on
+        changes to Ingress resources registered by the k8s api-server. This
+        means users of KubeIngressProxy may have greatly varying experiences of
+        using it depending on the performance of their k8s cluster setup.
+
+        Use of KubeIngressProxy as a JupyterHub Proxy class, is entirely
+        independent of use of KubeSpawner as a JupyterHub Spawner class. For
+        reasons related to sharing code with KubeSpawner in reflectors.py, it
+        has been made and remained part of the jupyterhub/kubespawner project.
+
+        Related issues:
+        - Need for tests: https://github.com/jupyterhub/kubespawner/issues/496
+        - Need for docs:  https://github.com/jupyterhub/kubespawner/issues/163
+
+    ---
+
+    KubeIngressProxy is an implementation of the JupyterHub Proxy class that
+    JupyterHub can be configured to rely on:
+
+        c.JupyterHub.proxy_class = "kubespawner:KubeIngressProxy"
+
+    Like all JupyterHub Proxy implementations, KubeIngressProxy will know
+    how to respond to hub requests like `get_all_routes`, `add_route`, and
+    `delete_route` in a way that will ensure traffic will get routed to the user
+    pods or JupyterHub registered external services. For reference, the official
+    documentation on writing a custom Proxy class like this is documented here:
+    https://jupyterhub.readthedocs.io/en/stable/reference/proxy.html.
+
+    KubeIngressProxy communicates with the k8s api-server in order to
+    create/delete Ingress resources according to the hub's `add_route`/`delete_route`
+    requests. It doesn't route traffic by itself, but instead relies on the k8s cluster's
+    ability to route traffic according to these Ingress resources.
+
+    Because KubeIngressProxy interacts with a k8s api-server that manages
+    Ingress resources, it must have permissions to do so as well. These
+    permissions should be granted to a k8s service account for where the
+    JupyterHub is running, as that is also where the KubeIngressProxy class
+    instance will run its code.
+
+    FIXME: Verify what k8s RBAC permissions are required for KubeIngressProxy
+           to function.
+
+           - The IngressReflector, ServiceReflector, and EndpointsReflector
+             require permission to read/list/watch those resources.
+           - `add_route` and `delete_route` requires permission to
+             create/update/patch/delete Ingress, Service, and Endpoints
+             resources.
+
+           These permissions are needed on a k8s Role resource bound to the k8s
+           ServiceAccount (via a k8s RoleBinding) used on the k8s Pod where
+           JupyterHub runs:
+
+           ```yaml
+           kind: Role
+           apiVersion: rbac.authorization.k8s.io/v1
+           metadata:
+           name: kube-ingress-proxy
+           rules:
+             - apiGroups: [""]
+               resources: ["endpoints", "services"]
+               verbs: ["get", "watch", "list", "create", "update", "patch", "delete"]
+             - apiGroups: ["networking.k8s.io"]
+               resources: ["ingresses"]
+               verbs: ["get", "watch", "list", "create", "update", "patch", "delete"]
+           ```
+    """
+
     namespace = Unicode(
         config=True,
         help="""
@@ -82,10 +165,10 @@ class KubeIngressProxy(Proxy):
         config=True,
         help="""
         Location (absolute filepath) for CA certs of the k8s API server.
-        
-        Typically this is unnecessary, CA certs are picked up by 
+
+        Typically this is unnecessary, CA certs are picked up by
         config.load_incluster_config() or config.load_kube_config.
-        
+
         In rare non-standard cases, such as using custom intermediate CA
         for your cluster, you may need to mount root CA's elsewhere in
         your Pod/Container and point this variable to that filepath
@@ -97,23 +180,15 @@ class KubeIngressProxy(Proxy):
         config=True,
         help="""
         Full host name of the k8s API server ("https://hostname:port").
-        
-        Typically this is unnecessary, the hostname is picked up by 
+
+        Typically this is unnecessary, the hostname is picked up by
         config.load_incluster_config() or config.load_kube_config.
         """,
     )
 
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
-
-        # We use the maximum number of concurrent user server starts (and thus proxy adds)
-        # as our threadpool maximum. This ensures that contention here does not become
-        # an accidental bottleneck. Since we serialize our create operations, we only
-        # need 1x concurrent_spawn_limit, not 3x.
-        self.executor = ThreadPoolExecutor(max_workers=self.app.concurrent_spawn_limit)
-
-        # Global configuration before reflector.py code runs
-        self._set_k8s_client_configuration()
+        load_config(host=self.k8s_api_host, ssl_ca_cert=self.k8s_api_ssl_ca_cert)
         self.core_api = shared_client('CoreV1Api')
         self.extension_api = shared_client('ExtensionsV1beta1Api')
 
@@ -128,34 +203,19 @@ class KubeIngressProxy(Proxy):
             parent=self, namespace=self.namespace, labels=labels
         )
         self.endpoint_reflector = EndpointsReflector(
-            parent=self, namespace=self.namespace, labels=labels
+            self, namespace=self.namespace, labels=labels
         )
 
-    def _set_k8s_client_configuration(self):
-        # The actual (singleton) Kubernetes client will be created
-        # in clients.py shared_client but the configuration
-        # for token / ca_cert / k8s api host is set globally
-        # in kubernetes.py syntax.  It is being set here
-        # and this method called prior to shared_client
-        # for readability / coupling with traitlets values
-        try:
-            kubernetes.config.load_incluster_config()
-        except kubernetes.config.ConfigException:
-            kubernetes.config.load_kube_config()
-        if self.k8s_api_ssl_ca_cert:
-            global_conf = client.Configuration.get_default_copy()
-            global_conf.ssl_ca_cert = self.k8s_api_ssl_ca_cert
-            client.Configuration.set_default(global_conf)
-        if self.k8s_api_host:
-            global_conf = client.Configuration.get_default_copy()
-            global_conf.host = self.k8s_api_host
-            client.Configuration.set_default(global_conf)
+        # schedule our reflectors to start in the event loop,
+        # reflectors first load can be awaited with:
+        #
+        #   await some_reflector._first_load_future
+        #
+        asyncio.ensure_future(self.ingress_reflector.start())
+        asyncio.ensure_future(self.service_reflector.start())
+        asyncio.ensure_future(self.endpoint_reflector.start())
 
-    @run_on_executor
-    def asynchronize(self, method, *args, **kwargs):
-        return method(*args, **kwargs)
-
-    def safe_name_for_routespec(self, routespec):
+    def _safe_name_for_routespec(self, routespec):
         safe_chars = set(string.ascii_lowercase + string.digits)
         safe_name = generate_hashed_slug(
             'jupyter-'
@@ -164,7 +224,7 @@ class KubeIngressProxy(Proxy):
         )
         return safe_name
 
-    async def delete_if_exists(self, kind, safe_name, future):
+    async def _delete_if_exists(self, kind, safe_name, future):
         try:
             await future
             self.log.info('Deleted %s/%s', kind, safe_name)
@@ -177,7 +237,8 @@ class KubeIngressProxy(Proxy):
         # Create a route with the name being escaped routespec
         # Use full routespec in label
         # 'data' is JSON encoded and put in an annotation - we don't need to query for it
-        safe_name = self.safe_name_for_routespec(routespec).lower()
+
+        safe_name = self._safe_name_for_routespec(routespec).lower()
         labels = {
             'heritage': 'jupyterhub',
             'component': self.component_label,
@@ -189,9 +250,7 @@ class KubeIngressProxy(Proxy):
 
         async def ensure_object(create_func, patch_func, body, kind):
             try:
-                resp = await self.asynchronize(
-                    create_func, namespace=self.namespace, body=body
-                )
+                await create_func(namespace=self.namespace, body=body)
                 self.log.info('Created %s/%s', kind, safe_name)
             except client.rest.ApiException as e:
                 if e.status == 409:
@@ -199,8 +258,7 @@ class KubeIngressProxy(Proxy):
                     self.log.warn(
                         "Trying to patch %s/%s, it already exists", kind, safe_name
                     )
-                    resp = await self.asynchronize(
-                        patch_func,
+                    await patch_func(
                         namespace=self.namespace,
                         body=body,
                         name=body.metadata.name,
@@ -222,13 +280,12 @@ class KubeIngressProxy(Proxy):
                 'Could not find endpoints/%s after creating it' % safe_name,
             )
         else:
-            delete_endpoint = self.asynchronize(
-                self.core_api.delete_namespaced_endpoints,
+            delete_endpoint = await self.core_api.delete_namespaced_endpoints(
                 name=safe_name,
                 namespace=self.namespace,
                 body=client.V1DeleteOptions(grace_period_seconds=0),
             )
-            await self.delete_if_exists('endpoint', safe_name, delete_endpoint)
+            await self._delete_if_exists('endpoint', safe_name, delete_endpoint)
 
         await ensure_object(
             self.core_api.create_namespaced_service,
@@ -260,26 +317,24 @@ class KubeIngressProxy(Proxy):
         # We just ensure that these objects are deleted.
         # This means if some of them are already deleted, we just let it
         # be.
-        safe_name = self.safe_name_for_routespec(routespec).lower()
+
+        safe_name = self._safe_name_for_routespec(routespec).lower()
 
         delete_options = client.V1DeleteOptions(grace_period_seconds=0)
 
-        delete_endpoint = self.asynchronize(
-            self.core_api.delete_namespaced_endpoints,
+        delete_endpoint = await self.core_api.delete_namespaced_endpoints(
             name=safe_name,
             namespace=self.namespace,
             body=delete_options,
         )
 
-        delete_service = self.asynchronize(
-            self.core_api.delete_namespaced_service,
+        delete_service = await self.core_api.delete_namespaced_service(
             name=safe_name,
             namespace=self.namespace,
             body=delete_options,
         )
 
-        delete_ingress = self.asynchronize(
-            self.extension_api.delete_namespaced_ingress,
+        delete_ingress = await self.extension_api.delete_namespaced_ingress(
             name=safe_name,
             namespace=self.namespace,
             body=delete_options,
@@ -293,15 +348,16 @@ class KubeIngressProxy(Proxy):
         # explicitly ourselves as well. In the future, we can probably try a
         # foreground cascading deletion (https://kubernetes.io/docs/concepts/workloads/controllers/garbage-collection/#foreground-cascading-deletion)
         # instead, but for now this works well enough.
-        await self.delete_if_exists('endpoint', safe_name, delete_endpoint)
-        await self.delete_if_exists('service', safe_name, delete_service)
-        await self.delete_if_exists('ingress', safe_name, delete_ingress)
+        await asyncio.gather(
+            self._delete_if_exists('endpoint', safe_name, delete_endpoint),
+            self._delete_if_exists('service', safe_name, delete_service),
+            self._delete_if_exists('ingress', safe_name, delete_ingress),
+        )
 
     async def get_all_routes(self):
-        # copy everything, because iterating over this directly is not threadsafe
-        # FIXME: is this performance intensive? It could be! Measure?
-        # FIXME: Validate that this shallow copy *is* thread safe
-        ingress_copy = dict(self.ingress_reflector.ingresses)
+        if not self.ingress_reflector.first_load_future.done():
+            await self.ingress_reflector.first_load_future
+
         routes = {
             ingress["metadata"]["annotations"]['hub.jupyter.org/proxy-routespec']: {
                 'routespec': ingress["metadata"]["annotations"][
@@ -314,7 +370,7 @@ class KubeIngressProxy(Proxy):
                     ingress["metadata"]["annotations"]['hub.jupyter.org/proxy-data']
                 ),
             }
-            for ingress in ingress_copy.values()
+            for ingress in self.ingress_reflector.ingresses.values()
         }
 
         return routes
